class StackOutOfBoundsException extends RuntimeException {
	String message = new String("Stack is out of bounds!");
}

class AtariCPU implements ESCCodes, CPU_M6502 {
	public AtariMem memory;
	AtariDevice device_handlers = new AtariDevice();

	protected int PC, SP, flags;
	protected int A, X, Y;

	protected boolean IRQ;
	protected boolean NMI;
	protected boolean RST;

	public int counter = 0;
	public int instr = 0;

	boolean trace = true;
	boolean trace_jumps = false;
	boolean trace_instructions = false;

	int [] bits2bcd = new int [256];
	int [] bcd2bits = new int [256];


	/**
	 * Creates the new instance of our AtariCPU processor which
	 * directly extends the class M65C02 and adds the escape methods.
	 *
	 * @param   mem  memory object on what this CPU operates.
	 */
	public AtariCPU(AtariMem mem) {
		memory = mem;
		System.out.println("memory assigned");
		System.out.println("mem:"+memory.mem[560]);

		for (int i = 0; i < 256 ; i++) {
			bits2bcd[i] = ((i >> 4) & 0xf) * 10 + (i & 0xf);
			bcd2bits[i] = (((i % 100) / 10) << 4) | (i % 10);
		}

		cpuReset();
		debug();
	}


	/**
	 * CPU reset method. This method resets the CPU as we can found it
	 * after power up. It is: X=A=Y=0, IRQ and NMI aren't set, stack pointer
	 * SP is set to the top of the second page (SP = 0x1ff) and PC contains
	 * the address hold in 0xfffc and 0xfffd.
	 *
	 */
	public void cpuReset() {
		setA(0);
		setX(0);
		setY(0);
		IRQ = NMI = false;
		PC = resetVector();
		setSP (0x1ff);
		setFlags (NUSED);
	}

	/**
	 * Returns the actual CPU IRQ handler pointer. The Mostek 65C02 processor has its irq,
	 * nmi and reset vectors stored from addresses $fffa - $ffff. 
	 *
	 * @return   pointer to the irq handler routine.  
	 */
	public int irqVector() {
		return (memory.getWord(0xfffe));
	}

	/**
	 * Returns the actual CPU RESET handler pointer. The Mostek 65C02 processor has its irq,
	 * nmi and reset vectors stored from addresses $fffa - $ffff. RESET signal is active
	 * after power up.
	 *
	 * @return   pointer to the reset handler routine.  
	 */

	public int resetVector() {
		return (memory.getWord(0xfffc));
	}

	/**
	 * Returns the actual CPU NMI handler pointer. The Mostek 65C02 processor has its irq,
	 * nmi and reset vectors stored from addresses $fffa - $ffff. NMI signal is caused
	 * by the Antic CPU.
	 *
	 * @return   pointer to the NMI handler routine.  
	 */

	public int nmiVector() {
		return (memory.getWord(0xfffa));
	}



	/**
	 * Safe decrementation of SP (stack pointer) register.
	 * NOTE: I don't know if this routine works right.
	 *
	 */
	public void decSP() {
		SP--;
		if (SP<0x100) throw new StackOutOfBoundsException();
	}


	/**
	 * Safe incrementation of SP (stack pointer) register.
	 * NOTE: I don't know if this routine works right.
	 *
	 */
	public void incSP() {
		SP++;
		if (SP>0x1ff) throw new StackOutOfBoundsException();

	}


	/**
	 * Pushes the value onto the CPU stack. It also decrements
	 * the SP (stack pointer) register using safe decrement
	 * decSP() method. This method throws the StackOfBoundsException when
	 * you exceed the stack range.
	 *
	 * @param   value the value to push onto the stack. 
	 */
	public void push(int value) {
		memory.mem[SP] = (value & 0xff);


		try {
			decSP();
		}
		catch (StackOutOfBoundsException e) {	
			debug();
			System.exit(1);
		}

	}


	/**
	 * Pops the value from the CPU stack. It also increments
	 * the SP (stack pointer) register using safe increment incSP()
	 * method. This method throws the StackOfBoundsException when
	 * you exceed the stack range.
	 *
	 * @return     returns the popped value from the stack.
	 */
	public int pop() {
		try {
			incSP();
		}
		catch (StackOutOfBoundsException e) {	
			debug();
			System.exit(1);
		}

		return memory.mem[SP] & 0xff;
	}


	/**
	 * Loads value from memory to A (accumulator) register. We
	 * use this in decodeInstruction() method.
	 *
	 * @param   memptr  points to memory from where to load the value
	 */
	public void loadValueToA(int memptr) {
		A = memory.getByte( memptr );
	}


	/**
	 *  Loads value from memory to X (index) register. We
	 * use this in decodeInstruction() method.
	 *
	 * @param   memptr  points to memory from where to load the value
	 */
	public void loadValueToX(int memptr) {
		X = memory.getByte( memptr );
	}


	/**
	 * Loads value from memory to Y (index) register. We
	 * use this in decodeInstruction() method.
	 *
	 * @param   memptr  points to memory from where to load the value

*/
	public void loadValueToY(int memptr) {
		Y = memory.getByte( memptr );
	}


	/**
	 * Stores value from A (accumulator) register into the memory. We
	 * use this in decodeInstruction() method.
	 *
	 * @param   memptr   points to the memory where to store the value

*/
	public void storeValueFromA( int memptr ) {
		memory.putByte (memptr, A);
	}


	/**
	 * Stores value from X (index) register into the memory. We
	 * use this in decodeInstruction() method.
	 *
	 * @param   memptr  points to the memory where to store the value

*/
	public void storeValueFromX( int memptr ) {
		memory.putByte (memptr, X);
	}


	/**
	 * Stores value from Y (index) register into the memory. We
	 * use this in decodeInstruction() method.
	 *
	 * @param   memptr  points to the memory where to store the value

*/
	public void storeValueFromY( int memptr ) {
		memory.putByte (memptr, Y);
	}

	/**
	 * Only for debug purposes. It displays the actual contents of all 
	 * CPU registers.
	 */

	public void debug() {
		System.out.println("***Debug info (instr="+instr+"):");
		System.out.println("           A = " + A + "   X = " + X + "     Y = " + Y);
		System.out.println("          SP = " + SP+ "  PC = " + PC+ " flags = " + flags);
		System.out.println("          IRQ = " + irqVector() + "  RES = " + resetVector() + "  NMI = " + nmiVector());
		System.out.println(" memory[pc]="+memory.getByte(PC));
	}


	/**
	 *		When an attempt to execute illegal instruction comes we halt the
	 *		cpu. An infinite loop will be taken.
	 */
	protected 	void halt() {
		while (true) ;
	}


	/**
	 * It simply initializes the Accumulator register (A) of Mostek 65C02 CPU
	 *
	 * @param   A  value to pass to the accumulator
	 */

	public void setA(int A) {
		this.A = A & 0xff;
	}


	/**
	 *		This method simply load the passed value into the X index register
	 *   of Mostek 65C02 CPU.
	 *
	 * @param   X  value to be stored into the X register
	 */
	public void setX(int X) {
		this.X = X & 0xff;
	}

	/**
	 *		This method simply load the passed value into the Y index register
	 *   of Mostek 65C02 CPU.
	 *
	 * @param   Y  value to be stored into the Y register
	 */

	public void setY(int Y) {
		this.Y = Y & 0xff;
	}


	/**
	 * The explicit set of Program Counter (PC). The PC points to the memory
	 * from the CPU executes instructions
	 *
	 * @param   PC  the value sets the Program Counter to the place in memory
	 */
	public void setPC(int PC) {
		this.PC = PC & 0xffff;
	}


	/**
	 * An explicit set of Stack Pointer (SP). This is space in memory where
	 * the CPU stack is located. On Mostek 65C02 it is the location $1ff as
	 * a top of the stack. When SP is $100 the stack is full.
	 *
	 * @param   SP  explicit set of stack pointer.
	 */
	public void setSP(int val) {
		if ( (val>0xff) && (val<0x200) )
			SP = val;
	}


	/**
	 * The explicit flags setting.
	 *
	 * @param   flags  binary value to set the flags register
	 */
	public void setFlags(int flags) {
		this.flags = flags & 0xff;
	}



	/**
	 * Immidiate addressing mode. We use this method to decode the
	 * effective address 
	 *
	 * @return     returns the immidiate address
	 */

	protected 	int imm() {
		return PC + 1;
	}


	/**
	 *	 Absolute addressing method. We use this method to decode the
	 * effective address 
	 *
	 * @return     returns the absolute address 
	 */
	protected 	int	abs() {
		int hi_byte, low_byte;

		try {
			hi_byte = memory.mem[PC + 2];
			low_byte = memory.mem[PC + 1];
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("Invalid absolute address has occured.\nYour computer has crashed.");
			hi_byte = low_byte = 0x0000;
			halt();
		}

		return (hi_byte << 8) | low_byte;
	}

	/*
	 * what if abs is 0xfffc and X = 0xff? What is the result ?!!
	 * Probably it is 0xffff...
	 */

	/**
	 * Absolute X indexed addressing. We use this method to decode the
	 * effective address 
	 *
	 * @return     returns the absolute x-indexed address.
	 */
	protected		int	absX() {
		int hi_byte, low_byte;

		try {
			hi_byte = memory.mem[PC + 2];
			low_byte = memory.mem[PC + 1];
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("Invalid absolute address has occured.\nYour computer has crashed.");
			hi_byte = low_byte = 0x0000;
			halt();
		}

		return (hi_byte << 8) + low_byte + X; 
	}


	/**
	 * Absolute Y-index addresing. We use this method to decode the
	 * effective address 
	 *
	 * @return     returns absolute Y-indexed address
	 */
	protected		int	absY() {
		int hi_byte, low_byte;

		try {
			hi_byte = memory.mem[PC + 2];
			low_byte = memory.mem[PC + 1];
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("Invalid absolute address has occured.\nYour computer has crashed.");
			hi_byte = low_byte = 0x0000;
			halt();
		}

		return (hi_byte << 8) + low_byte + Y; 
	}


	/**
	 * Zero page addressing mode. This is special fast mode. But it can be
	 * used only in the first page from 0x0000 to 0x00ff.
	 *
	 * @return     effective address
	 */
	protected		int	zp() {
		return memory.mem[PC + 1]; 
	}

	/**
	 * Zero page X-indexed addressing mode. This is special fast mode. But it can be
	 * used only in the first page from 0x0000 to 0x00ff.
	 *
	 * @return     effective address

*/
	protected 	int zpX() {
		return (memory.mem[PC + 1] + X) & 0xff;
	}

	/**
	 * Zero page Y-indexed addressing mode. This is special fast mode. But it can be
	 * used only in the first page from 0x0000 to 0x00ff.
	 *
	 * @return     effective address

*/
	protected 	int zpY() {
		return (memory.mem[PC + 1] + Y) & 0xff;
	}


	/**
	 * Indirect X-indexed addressing mode. We use this method to decode the
	 * effective address 
	 *
	 * @return     returns the indirect effective address.
	 */
	protected int indX() {
		int low = (memory.mem[PC + 1] + X) & 0xff;
		return (memory.mem[low + 1] << 8) + memory.mem[low]; 
	}


	/**
	 * Indirect Y-indexed addressing mode. We use this method to decode the
	 * effective address 
	 *
	 * @return     returns the indirect effective address.
	 */
	protected int indY() {
		int low, ind ;

		low = memory.mem[PC + 1];
		ind = (memory.mem[low + 1] << 8) + memory.mem[low] + Y; 

		return ind;

	}

	public void testRegister(int reg, int FLAGS) {
		if ((FLAGS & ZER) != 0) { 
			if (reg==0) flags |= ZER;
			else   	flags  &= ~ZER;
		}

		if ((FLAGS & NEG) != 0) {
			if ((reg & 0x80) != 0) flags |= NEG;
			else           flags &= ~NEG;
		}
	}


	/**
	 * OBSOLETE.
	 * Adjust all registers to their bit width. We call this method after
	 * some (almost) instructions which could cause carry, etc., because we
	 * inernaly represent all CPU registres as int which is of course 4bytes
	 * long. So we detect if carry flag has to be set and than we adjust
	 * the register to  its appropriate bit width.
	 */
	public void repairAllRegisters() {
		A &= 0xff;
		X &= 0xff;
		Y &= 0xff;
		PC &= 0xffff;
		SP &= 0xffff;
		flags &= 0xff;
	}



	/**
	 * For internal use in method decodeInstruction().
	 *
	 * @param   source  operand to compare with destination
	 * @param   destination  
	 */
	protected void CMP(int source, int destination) {
		if ( source == destination ) {
			flags |= ZER;
		} else {
			flags &= ~ZER;
		}

		if ( source < destination ) {
			flags &= ~CAR;
		} else {
			flags |= CAR;
		}

		if ( ( (source - destination) >= 0 ) && ( (source - destination) <= 0x7f ) ) {
			flags &= ~NEG;
		}	 

		if ( ( (source - destination) >= 0x80 ) && ( (source - destination) <= 0xff ) ) {
			flags |= NEG;
		} 
	}


	protected void ADC(int op) {
		int old_A = A, bcd1, bcd2;

		if ((flags & DEC) == DEC) {
			//System.out.println("BCD");
			bcd1 = bits2bcd [old_A];
			bcd2 = bits2bcd [op];
			//System.out.println("bcd1="+bcd1+" (A="+A+")");
			//System.out.println("bcd2="+bcd2+" (op="+op+")");
			bcd1 += bcd2;
			bcd1 += (flags & CAR);
			if ( bcd1 > 99 ) {
				bcd1-=100;
				flags |= CAR;
			} else {
				flags &= ~CAR;
			}

			//System.out.println("bcd1="+bcd1);
			A = bcd2bits [bcd1];
			//System.out.println("A="+A);

			testRegister (A, NEG|ZER);
			A &= 0xff;
			old_A&= 0xff;
			//System.out.println("BCD ADC("+old_A+"+"+op+")="+A);
			if (((A ^ old_A) & 0x80) == 0x80) {
				flags |= OVR;
			} else {
				flags &= ~OVR;
			}

		} else {			// normal hexadecimal mode
			A += op & 0xff;
			A += (flags & CAR);			// CAR is defined as CAR = 1, ok.
			testRegister (A, NEG|ZER);	//sets NEG and ZERO flags

			if ( A > 0xff )				// carry
				flags |= CAR;		// set carry
			else
				flags &= ~CAR;

			A &= 0xff;
			old_A&= 0xff;
			if (((A ^ old_A) & 0x80) == 0x80)
				flags |= OVR;
			else
				flags &= ~OVR;

		}
	}


	protected void SBC(int op) {
		int old_A = A, bcd1, bcd2;
		int C;

		if ( (flags & CAR) != 0) 
			C = 0;			// negative CARRY : C = ~CAR & 0x1 (if CAR=1)
		else
			C = 1;

		if ((flags & DEC) == DEC) {
			// not implemented yet.
			bcd1 = bits2bcd [A];
			bcd2 = bits2bcd [op];

			bcd1 = bcd1 - bcd2;
			bcd1 -= C;

			if ( bcd1 < 0 ) {
				bcd1 = 100 - ( -bcd1 );
			}

			A = bcd2bits [bcd1];


			if (old_A < (op  + C) ) 
				flags &= ~CAR;
			else
				flags |= CAR;


			A    &= 0xff;
			old_A&= 0xff;
			if (((A ^ old_A) & 0x80) == 0x80)
				flags |= OVR;
			else
				flags &= ~OVR;

		} else {				// normal hexadecimal mode
			op &= 0xff;
			A -= op; 
			A -= C;				// CAR is defined as CAR = 1, ok.
			testRegister (A, NEG|ZER);		//sets NEG and ZERO flags

			if (old_A < (op  + C) ) 
				flags &= ~CAR;
			else
				flags |= CAR;


			A &= 0xff;
			old_A&= 0xff;
			if (((A ^ old_A) & 0x80) == 0x80)
				flags |= OVR;
			else
				flags &= ~OVR;

		}
	}

	protected int ASL(int reg) {
		int data;
		if  ( (reg & 0x80) == 0x80 )
			flags |= CAR;
		else
			flags &= ~CAR;


		data = reg << 1;
		data &=0xfe;
		testRegister(data, ZER|NEG);

		//				System.out.println("ASL("+reg+")="+data);
		return data;
	}

	protected int LSR(int reg) {
		int data;

		if ( (reg & 0x1) == 0x1 )
			flags |= CAR;
		else
			flags &= ~CAR;

		data = reg >>> 1;
		data &=0x7f;
		flags &= ~NEG;
		testRegister(reg, ZER);

		return data;
	}

	protected void JMP() {
		int where = memory.mem[++PC] & 0xff;

		PC++;

		if ( (where & 0x80) == 0 )
			PC += where;
		else
			PC += (where - 256);

		/*	if (PC!=49837)
			System.out.println("JMP(): PC="+PC);
			*/
	}

	protected int ROL(int reg) {
		int data;

		data = reg << 1;
		data &=0xfe;

		if ( (flags & CAR) != 0) 
			data |= 1;

		if ( ( reg & 0x80 ) != 0) 
			flags |= CAR;
		else
			flags &= ~CAR;

		testRegister(data, NEG|ZER);

		return data;
	}

	protected int ROR(int reg) {
		int data;

		data = reg;
		data>>=1;
		data &=0x7f;
		if ( (flags & CAR) != 0) {
			data |= 0x80;
		}

		if (( reg & 0x1 ) != 0)
			flags |= CAR;
		else
			flags &= ~CAR;

		testRegister(data, NEG|ZER);

//		System.out.println("ROR("+reg+")="+data);
		return data;
	}

	protected void BIT(int reg) {
		if ( (A & reg) == 0 )	
			flags |= ZER;
		else
			flags &= ~ZER;


		if ( (reg & 0x80) != 0 ) 
			flags |= NEG;
		else
			flags &= ~NEG;

		if ( (reg & 0x40) != 0 )
			flags |= OVR;
		else
			flags &= ~OVR;
	}


	protected int I_ABSJMP() {
		return memory.mem[PC+1] | (memory.mem[PC+2] << 8);
	}


	/**
	 * We use this method to add some "so as called" escape codes
	 * into CPU memory. Theese are not valid CPU instructions we
	 * use them only for emulation purposes such as modifying 
	 * the HATABS table. We hook to escape codes our routines.
	 *
	 * @param   esc_code		where to escape  
	 */
	public void escapeSyscall(int esc_code) {
		int [] AReg = new int[1];
		int [] YReg = new int[1];
		int [] FReg = new int[1];


		System.out.println("ESC_syscall : "+esc_code);

		switch (esc_code) {
			case ESC_SIO :
				// SIO (); not implemented yet
				System.out.println("SIO routine...");
				Y = 146;
				flags |= NEG;
				memory.putByte(0x303, Y);
				break;

			case ESC_K_OPEN :
			case ESC_K_CLOSE :
			case ESC_K_READ :
			case ESC_K_WRITE :
			case ESC_K_STATUS :
			case ESC_K_SPECIAL :
				/*device_handlers.keyboardDevice (esc_code, AReg, YReg, FReg);
				  A = AReg[0];
				  Y = YReg[0];
				  flags = FReg[0];*/
				break;
			case ESC_H_OPEN :
			case ESC_H_CLOSE :
			case ESC_H_READ :
			case ESC_H_WRITE :
			case ESC_H_STATUS :
			case ESC_H_SPECIAL :
				// H_Device (esc_code); not implmented yet
				break;
			case ESC_E_OPEN :
			case ESC_E_READ :
			case ESC_E_WRITE :
				device_handlers.editorDevice (esc_code, AReg, YReg, FReg);
				A = AReg[0];
				Y = YReg[0];
				flags = FReg[0];
				break;
			default :
				System.out.println("Invalid ESC Code "+ esc_code);
				System.exit (1);
		}

	}


	/**
	 *		This is the main CPU routine. It execute the instruction
	 * 	from the current PC pointer.
	 */
	public void decodeInstruction(int ncycles) {

		if (NMI) {
			push (PC >> 8);
			push (PC & 0xff);
			push (flags);
			flags |= INT;
			PC = nmiVector();
			NMI = false;
		}

		if (IRQ)	{
			//System.out.println("IRQ stage 1 !!!");

			//	if ((flags & INT) == 0)  {
			push (PC >> 8);
			push (PC & 0xff);
			push (flags);
			flags |= INT;
			PC = irqVector();
			IRQ = false;
			//	}
		}



		while (ncycles-- != 0) {							
			counter++;

			if (memory.mem[0xbffc] != 0) {
				System.out.println("0xbffc != 0 @ "+PC);
				System.exit(-1);
			}

			//						System.out.println(++instr+": PC = "+PC+" A="+A+" X="+X+" Y="+Y+" op="+memory.mem[PC]+" P="+flags);

			switch ( memory.mem[ PC ] ) {
				case 0xff:							// ESC!!!
					// This is not the valid 65C02 instruction.
					// We use it only for emulation purposes!

					escapeSyscall(memory.mem[PC + 1]);
					PC += 2;			// should be RTS
					ncycles = 1;
					break;

				case 0x69:							// ADC #op8
					ADC( memory.getByte( imm() ));
					PC += 2;
					break;

				case 0x65:							// ADC $op8
					ADC( memory.getByte( zp() ));
					PC += 2;
					break;

				case 0x75: 						// ADC $op8, X
					ADC( memory.getByte( zpX() ));
					PC += 2;
					break;

				case 0x6d:							// ADC $op16
					ADC( memory.getByte( abs() ));
					PC += 3;
					break;

				case 0x7d:							// ADC $op16, X
					ADC( memory.getByte( absX() ));
					PC += 3;
					break;

				case 0x79:							// ADC $op16, Y
					ADC( memory.getByte( absY() ));
					PC += 3;
					break;

				case 0x61:							// ADC ($op8, X)
					ADC( memory.getByte( indX() ));
					PC += 2;
					break;

				case 0x71:							// ADC ($op8), Y
					ADC( memory.getByte( indY() ));
					PC += 2;
					break;

				case 0x29:							// AND #op8
					A &= memory.getByte( imm() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x25:							// AND $op8
					A &= memory.getByte( zp() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x35:							// AND $op8, X
					A &= memory.getByte( zpX() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x2d:							// AND $op16
					A &= memory.getByte( abs() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x3d:							// AND $op16, X
					A &= memory.getByte( absX() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x39:							// AND $op16, Y
					A &= memory.getByte( absY() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x21:							// AND ($op8, X)
					A &= memory.getByte( indX() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x31:							// AND ($op8), Y
					A &= memory.getByte( indY() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x0a:							// ASL A
					A = ASL (A);
					PC++;
					break;

				case 0x06:							// ASL $op8
					memory.putByte ( zp(), ASL (memory.getByte(zp()) ) );
					PC += 2;
					break;

				case 0x16:							// ASL $op8, X
					memory.putByte ( zpX(), ASL (memory.getByte(zpX()) ) );
					PC += 2;							
					break;

				case 0x0e:							// ASL $op16
					memory.putByte ( abs(), ASL (memory.getByte(abs()) ) );
					PC += 3;
					break;

				case 0x1e:							// ASL $op16, X
					memory.putByte ( absX(), ASL (memory.getByte(absX()) ) );
					PC += 3;
					break;

				case 0x90:							// BCC $op8
					if ( (flags & CAR) == 0 )
						JMP ();
					else PC+=2;
					break;

				case 0xb0:							// BCS $op8
					if ( (flags & CAR) != 0 )
						JMP ();
					else PC+=2;
					break;

				case 0xf0:							// BEQ $op8
					if ( (flags & ZER) != 0 )
						JMP ();
					else PC+=2;
					break;

				case 0x24:							// BIT $op8
					BIT (memory.getByte(zp()));
					PC += 2;
					break;

				case 0x2c:							// BIT $op16
					BIT (memory.getByte(abs()));
					PC += 3;
					break;


				case 0x30:							// BMI $op8
					if ( (flags & NEG) != 0 )
						JMP ();
					else PC+=2;
					break;

				case 0xd0:							// BNE $op8
					if ( (flags & ZER) == 0 )
						JMP ();
					else PC+=2;
					break;

				case 0x10:							// BPL $op8
					if ( (flags & NEG) == 0 )
						JMP ();
					else PC+=2;
					break;

				case 0x00: 							// BRK		
					PC+=2;
					push ((PC & 0xff00) >> 8);		// higher byte
					push (PC & 0x00ff);		// lower byte
					push (flags);					// flags
					PC = irqVector();
					flags |= BRK;
					break;

				case 0x50:							// BVC $op8
					if ( (flags & OVR) == 0 )
						JMP ();
					else PC+=2;
					break;

				case 0x70:							// BVS $op8
					if ( (flags & OVR) != 0 )
						JMP ();
					else PC+=2;
					break;

				case 0x18:							// CLC
					flags &= ~CAR;
					PC++;
					//System.out.println("CLC");
					break;

				case 0xd8:							// CLD
					flags &= ~DEC;
					PC++;
					//System.out.println("CLD");
					break;

				case 0x58:							// CLI
					flags &= ~INT;
					PC++;
					//System.out.println("CLI");
					break;

				case 0xb8:							// CLV
					flags &= ~OVR;
					PC++;
					//System.out.println("CLV");
					break;

				case 0xc9:							// CMP #op8
					CMP( A, memory.getByte(imm()) );
					repairAllRegisters();
					//System.out.println("CMP #"+memory.mem[imm()]);
					PC += 2;		
					break;


				case 0xc5:							// CMP $op8
					CMP( A, memory.getByte(zp()) );
					repairAllRegisters();
					//System.out.println("CMP $"+memory.mem[zp()]);
					PC += 2;		
					break;

				case 0xd5:							// CMP #op8, X
					CMP( A, memory.getByte(zpX()) );
					repairAllRegisters();
					//System.out.println("CMP $"+memory.mem[zpX()]+", X");
					PC += 2;		
					break;


				case 0xcd:							// CMP $op16
					CMP( A, memory.getByte(abs()) );
					repairAllRegisters();
					//System.out.println("CMP $"+memory.mem[abs()]);
					PC += 3;		
					break;

				case 0xdd:							// CMP $op16, X
					CMP( A, memory.getByte(absX()) );
					repairAllRegisters();
					//System.out.println("CMP $"+memory.mem[absX()]+", X");
					PC += 3;		
					break;

				case 0xd9:							// CMP $op16, Y
					CMP( A, memory.getByte(absY()) );
					repairAllRegisters();
					//System.out.println("CMP $"+memory.mem[absY()]+", Y");
					PC += 3;		
					break;

				case 0xc1:							// CMP ($op8, X)
					CMP( A, memory.getByte(indX()) );
					repairAllRegisters();
					//System.out.println("CMP ($"+memory.mem[indX()]+", X)");
					PC += 2;		
					break;

				case 0xd1:							// CMP ($op8), Y
					CMP( A, memory.getByte(indY()) );
					repairAllRegisters();
					//System.out.println("CMP ($"+memory.mem[indY()]+"), Y");
					PC += 2;		
					break;

				case 0xe0:							// CPX #op8
					CMP( X, memory.getByte(imm()) );
					repairAllRegisters();
					//System.out.println("CPX #"+memory.mem[imm()]);
					PC += 2;		
					break;

				case 0xe4:							// CPX $op8
					CMP( X, memory.getByte(zp()) );
					repairAllRegisters();
					PC += 2;		
					break;

				case 0xec:							// CPX $op16
					CMP( X, memory.getByte(abs()) );
					repairAllRegisters();
					PC += 3;		
					break;

				case 0xc0:							// CPY #op8
					CMP( Y, memory.getByte(imm()) );
					repairAllRegisters();
					PC += 2;		
					break;

				case 0xc4:							// CPY $op8
					CMP( Y, memory.getByte(zp()) );
					repairAllRegisters();
					PC += 2;		
					break;

				case 0xcc:							// CPY $op16
					CMP( Y, memory.getByte(abs()) );
					repairAllRegisters();
					PC += 3;		
					break;

				case 0xc6:							// DEC $op8
					memory.putByte( zp(), memory.getByte(zp()) - 1);
					testRegister(memory.getByte(zp()), NEG|ZER);
					PC += 2;
					break;

				case 0xd6:							// DEC $op8, X
					memory.putByte( zpX(), memory.getByte(zpX()) - 1);
					testRegister(memory.getByte(zpX()), NEG|ZER);
					PC += 2;
					break;

				case 0xce:							// DEC $op16
					memory.putByte( abs(), memory.getByte(abs()) - 1);
					testRegister(memory.getByte(abs()), NEG|ZER);
					PC += 3;
					break;

				case 0xde:							// DEC $op16, X
					memory.putByte( absX(), memory.getByte(absX()) - 1);
					testRegister(memory.getByte(absX()), NEG|ZER);
					PC += 3;
					break;


				case 0xca:							// DEX
					X--;							// THESE SHOULD BE ATOMIC!
					X &= 0xff;

					testRegister(X, NEG|ZER);
					//						repairAllRegisters();
					PC++;
					break;


				case 0x88:							// DEY
					Y--;	
					Y &= 0xff;

					testRegister(Y, NEG|ZER);
					PC++;
					break;

				case 0x49:							// EOR #op8
					A ^= memory.getByte( imm() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x45:							// EOR $op8
					A ^= memory.getByte( zp() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x55:							// EOR $op8, X
					A ^= memory.getByte( zpX() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x4d:							// EOR $op16
					A ^= memory.getByte( abs() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x5d:							// EOR $op16, X
					A ^= memory.getByte( absX() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x59:							// EOR $op16, Y
					A ^= memory.getByte( absY() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x41:							// EOR ($op8, X)
					A ^= memory.getByte( indX() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x51:							// EOR ($op8), Y
					A ^= memory.getByte( indY() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;


				case 0xe6:							// INC $op8
					memory.putByte ( zp(), memory.getByte(zp()) + 1);
					testRegister(memory.getByte(zp()), NEG|ZER);
					PC += 2;
					break;

				case 0xf6:							// INC $op8, X
					memory.putByte ( zpX(), memory.getByte(zpX()) + 1);
					testRegister(memory.getByte(zpX()), NEG|ZER);
					PC += 2;
					break;

				case 0xee:							// INC $op16
					memory.putByte ( abs(), memory.getByte(abs()) + 1);
					testRegister(memory.getByte(abs()), NEG|ZER);
					PC += 3;
					break;

				case 0xfe:							// INC $op16, X
					memory.putByte ( absX(), memory.getByte(absX()) + 1);
					testRegister(memory.getByte(absX()), NEG|ZER);
					PC += 3;
					break;

				case 0xe8:							// INX
					X++;			 // THESE *SHOULD BE* ATOMIC! */
					X &= 0xff;//
					testRegister(X, NEG|ZER);
					//System.out.println("X="+X+" Zero="+((flags & ZER) !=0));
					PC++;
					break;

				case 0xc8:							// INY
					Y++;
					Y &= 0xff;
					testRegister(Y, NEG|ZER);
					PC++;
					break;

				case 0x4c:							//	 JMP #op16
					if (trace_jumps) {
						System.out.println("JMP from: "+PC+" to :"+(memory.mem[PC+1] | (memory.mem[PC+2] << 8) ));
					}
					/*System.out.println("NV1BDIZC");
					  if ((flags & NEG) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & OVR) != 0)
					  System.out.print("11");
					  else 
					  System.out.print("01");

					  if ((flags & BRK) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & DEC) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & INT) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & ZER) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & CAR) != 0)
					  System.out.println("1");
					  else 
					  System.out.println("0");
					  */
					setPC (memory.mem[PC+1] | (memory.mem[PC+2] << 8) );
					break;

				case 0x6c:							//	 JMP (#op16)
					setPC (memory.mem[I_ABSJMP()] | (memory.mem[I_ABSJMP() + 1] << 8) );
					break;

				case 0x20:							// JSR #op16
					PC+=2;
					push ((PC & 0xff00) >> 8);		// PCH -> top
					push (PC & 0x00ff);					// PCL -> top

					if (trace_jumps) {
						System.out.println("JSR from :"+(PC-1)+" to:"+((memory.mem[PC]<<8) | memory.mem[PC - 1]) );
					}


					//if ((memory.mem[PC]<<8 | memory.mem[PC - 1]) == 62205) trace = true;
					/*System.out.println("NV1BDIZC");

					  if ((flags & NEG) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & OVR) != 0)
					  System.out.print("11");
					  else 
					  System.out.print("01");

					  if ((flags & BRK) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & DEC) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & INT) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & ZER) != 0)
					  System.out.print("1");
					  else 
					  System.out.print("0");

					  if ((flags & CAR) != 0)
					  System.out.println("1");
					  else 
					  System.out.println("0");
					  */
					setPC ((memory.mem[PC]<<8) | memory.mem[PC - 1]);
					//							System.out.println("PC="+PC);
					break;

				case 0xa9:							// LDA #op8
					loadValueToA ( imm() );
					testRegister(A, NEG|ZER);
					// repairAllRegisters();
					PC += 2;
					//								System.out.println("LDA #" + A);
					//								System.out.println("<55890> = " + memory.mem[55890]);
					break;

				case 0xa5:							// LDA $op8
					loadValueToA ( zp() );
					testRegister(A, NEG|ZER);
					//repairAllRegisters();
					//System.out.println("LDA $" + zp());
					PC += 2;
					break;

				case 0xb5:							// LDA $op8,x
					loadValueToA ( zpX() );
					testRegister(A, NEG|ZER);
					//repairAllRegisters();
					PC += 2;
					break;

				case 0xad:							// LDA $op16
					//System.out.println("LDA $"+abs());
					loadValueToA ( abs() );
					testRegister(A, NEG|ZER);
					//repairAllRegisters();
					PC += 3;
					break;

				case 0xbd:							// LDA $op16,X
					loadValueToA ( absX() );
					testRegister(A, NEG|ZER);
					repairAllRegisters();
					PC += 3;
					break;

				case	0xb9:						// LDA $op16, Y
					loadValueToA ( absY() );
					testRegister(A, NEG|ZER);
					//repairAllRegisters();
					PC += 3;
					break;


				case 0xa1:						// LDA ($op8, X)
					loadValueToA ( indX() );
					testRegister(A, NEG|ZER);
					//repairAllRegisters();
					PC += 2;
					break;

				case 0xb1:						// LDA ($op8), Y)
					//System.out.println("LDA ($op8), "+Y+" At "+PC+"  $5="+memory.mem[5]);
					loadValueToA ( indY() );
					testRegister(A, NEG|ZER);
					//repairAllRegisters();
					PC += 2;
					break;

				case 0xa2:						// LDX #op8
					loadValueToX ( imm() );
					testRegister(X, NEG|ZER);
					//System.out.println("LDX "+memory.mem[PC+1]);
					PC += 2;
					break;

				case 0xa6:						// LDX $op8
					loadValueToX ( zp() );
					testRegister(X, NEG|ZER);
					PC += 2;
					break;

				case 0xb6:						// LDX $op8, Y
					loadValueToX ( zpY() );
					testRegister(X, NEG|ZER);
					PC += 2;
					break;

				case 0xae:						// LDX $op16
					loadValueToX ( abs() );
					testRegister(X, NEG|ZER);
					PC += 3;
					break;

				case 0xbe:						// LDX $op16, Y
					loadValueToX ( absY() );
					testRegister(X, NEG|ZER);
					PC += 3;
					break;


				case 0xa0:						// LDY #op8
					loadValueToY ( imm() );
					testRegister(Y, NEG|ZER);
					PC += 2;
					break;

				case 0xa4:						// LDY $op8
					loadValueToY ( zp() );
					testRegister(Y, NEG|ZER);
					PC += 2;
					break;

				case 0xb4:						// LDY $op8, X
					loadValueToY ( zpX() );
					testRegister(Y, NEG|ZER);
					PC += 2;
					break;

				case 0xac:						// LDY $op16
					loadValueToY ( abs() );
					testRegister(Y, NEG|ZER);
					PC += 3;
					break;

				case 0xbc:						// LDY $op16, X
					loadValueToY ( absX() );
					testRegister(Y, NEG|ZER);
					PC += 3;
					break;

				case 0x4a:							// LSR A
					A = LSR (A);
					PC++;
					break;

				case 0x46:							// LSR $op8
					memory.putByte(zp(), LSR (memory.getByte(zp())));
					PC += 2;
					break;

				case 0x56:							// LSR $op8, X
					memory.putByte(zpX(), LSR (memory.getByte(zpX())));
					PC += 2;							
					break;

				case 0x4e:							// LSR $op16
					memory.putByte(abs(), LSR (memory.getByte(abs())));
					PC += 3;
					break;

				case 0x5e:							// LSR $op16, X
					memory.putByte(absX(), LSR (memory.getByte(absX())));
					PC += 3;
					break;


				case 0xea:						// NOP
					PC++;
					break;

				case 0x09:							// ORA #op8
					A |= memory.getByte( imm() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x05:							// ORA $op8
					A |= memory.getByte( zp() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x15:							// ORA $op8, X
					A |= memory.getByte( zpX() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x0d:							// ORA $op16
					A |= memory.getByte( abs() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x1d:							// ORA $op16, X
					A |= memory.getByte( absX() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x19:							// ORA $op16, Y
					A |= memory.getByte( absY() );
					testRegister (A, NEG|ZER);
					PC += 3;
					break;

				case 0x01:							// ORA ($op8, X)
					A |= memory.getByte( indX() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x11:							// ORA ($op8), Y
					A |= memory.getByte( indY() );
					testRegister (A, NEG|ZER);
					PC += 2;
					break;

				case 0x48:					// PHA
					push (A);
					PC++;
					break;

				case 0x08:					// PHP
					push (flags);
					PC++;
					break;

				case 0x68:					// PLA
					A = pop ();
					testRegister(A, NEG|ZER);
					PC++;
					break;

				case 0x28:					// PLP
					flags = pop();
					PC++;
					break;

				case 0x2a:					// ROL A
					A = ROL (A);
					PC++;
					break;

				case 0x26:					// ROL $op8
					memory.putByte(zp(), ROL (memory.getByte(zp())));
					PC += 2;
					break;

				case 0x36:					// ROL $op8, X
					memory.putByte(zpX(), ROL (memory.getByte(zpX())));
					PC += 2;
					break;

				case 0x2e:					// ROL $op16
					memory.putByte(abs(), ROL (memory.getByte(abs())));
					PC += 3;
					break;

				case 0x3e:					// ROL $op16, X
					memory.putByte(absX(), ROL (memory.getByte(absX())));
					PC += 3;
					break;


				case 0x6a:					// ROR A
					A = ROR (A);
					PC++;
					break;

				case 0x66:					// ROR $op8
					memory.putByte(zp(), ROR (memory.getByte(zp())));
					PC += 2;
					break;

				case 0x76:					// ROR $op8, X
					memory.putByte(zpX(), ROR (memory.getByte(zpX())));
					PC += 2;
					break;

				case 0x6e:					// ROR $op16
					memory.putByte(abs(), ROR (memory.getByte(abs())));
					PC += 3;
					break;

				case 0x7e:					// ROR $op16, X
					memory.putByte(absX(), ROR (memory.getByte(absX())));
					PC += 3;
					break;


				case 0x40:					// RTI
					flags = pop();
					PC = pop();
					PC |= pop() << 8;
					break;

				case 0x60:					// RTS
					PC = pop();
					PC |= pop() << 8;
					PC++;

					/*if (PC==50940) {
					  trace=true;
					  debug();
					  }*/
					//if (PC==49423) trace=false;

					//if (PC==50165) debug();
					/*
					   if (PC==50090) {
					   for (int q=0;q<30;q++) 
					   System.out.print(" "+memory.getByte(memory.getWord(560)+q));

					   System.out.println("");			
					   }*/

					if (trace_jumps) {
						System.out.println("RTS:"+PC);
					}

					//if (PC==50165) {
					//flags = 161;
					//debug();
					//int dl = memory.mem[560] + 256*memory.mem[561];
					//for (int i=0; i<20; i++) {
					//		System.out.print(" "+memory.mem[dl+i]);
					//}
					//}

					//halt();
					break;

				case 0xe9:					// SBC #op8	
					SBC (memory.getByte(imm()));
					PC += 2;
					break;

				case 0xe5:					// SBC $op8	
					SBC (memory.getByte(zp()));
					PC += 2;
					break;

				case 0xf5:					// SBC $op8, X
					SBC (memory.getByte(zpX()));
					PC += 2;
					break;

				case 0xed:					// SBC $op16
					SBC (memory.getByte(abs()));
					PC += 3;
					break;

				case 0xfd:					// SBC $op16, X
					SBC (memory.getByte(absX()));
					PC += 3;
					break;

				case 0xf9:					// SBC $op16, Y
					SBC (memory.getByte(absY()));
					PC += 3;
					break;

				case 0xe1:					// SBC ($op8, X)
					SBC (memory.getByte(indX()));
					PC += 2;
					break;

				case 0xf1:					// SBC ($op8), Y
					SBC (memory.getByte(indY()));
					PC += 2;
					break;

				case 0x38:				// SEC
					flags |= CAR;
					PC++;
					break;

				case 0xf8:				// SED
					flags |= DEC;
					PC++;
					break;

				case 0x78:				// SEI
					flags |= INT;
					//System.out.println("SEI");
					PC++;
					break;

					/*							
												case 0x1c:				// SKW UNOFFICIAL! skip word
												PC += 2;
												break;
												*/						
				case 0x85:			// STA $op8
					storeValueFromA ( zp() );
					PC += 2;
					break;

				case 0x95:			// STA $op8,X
					storeValueFromA ( zpX() );
					PC += 2;
					break;

				case 0x8d:			// STA $op16
					storeValueFromA ( abs() );
					PC += 3;
					break;

				case 0x9d:			// STA $op16,X
					storeValueFromA ( absX() );
					PC += 3;
					break;

				case 0x99:			// STA $op16,Y
					storeValueFromA ( absY() );
					PC += 3;
					break;

				case 0x81:			// STA ($op8, X)
					storeValueFromA ( indX() );
					PC += 2;
					break;

				case 0x91:			// STA ($op8), Y
					storeValueFromA ( indY() );
					PC += 2;
					break;

				case 0x86:			// STX $op8
					storeValueFromX ( zp() );
					PC += 2;
					break;

				case 0x96:			// STX $op8, X
					storeValueFromX ( zpX() );
					PC += 2;
					break;

				case 0x8e:			// STX $op16
					storeValueFromX ( abs() );
					PC += 3;
					break;

				case 0x84:			// STY $op8
					storeValueFromY ( zp() );
					PC += 2;
					break;

				case 0x94:			// STY $op8, X
					storeValueFromY ( zpX() );
					PC += 2;
					break;

				case 0x8c:			// STY $op16
					storeValueFromY ( abs() );
					PC += 3;
					break;


				case 0xaa:				// TAX
					X = A;
					testRegister(A, NEG|ZER);
					PC++;
					break;

				case 0xa8:				// TAY
					Y = A;
					testRegister(A, NEG|ZER);
					PC++;
					break;

				case 0xba:				// TSX
					X = SP & 0xff;
					testRegister(X, NEG|ZER);
					repairAllRegisters();
					PC++;
					break;

				case 0x8a:				// TXA
					A = X;
					testRegister(A, NEG|ZER);
					PC++;
					break;

				case 0x9a:				// TXS
					SP = X + STACK_BASE;
					testRegister(X, NEG|ZER);
					repairAllRegisters();
					PC++;
					break;

				case 0x98:			  // TYA
					A = Y;
					testRegister(A, NEG|ZER);
					PC++;
					break;

				default:
					System.out.println("Illegal instruction : "+memory.mem[PC]+" @ "+PC );
					debug();
					System.out.println("instr="+counter);
					System.exit(1);
			}
		}
	}

}
